#include <stdio.h>
#include <string.h>

#include "sage++user.h"
#include "anls_att.h"
#include "detect_comm.h"

extern char * mixstring;
extern Flag in_procedure;

extern void add_coef(SgStatement * s, int counter, int * amount);
extern void create_section(SgStatement * s, int counter);
extern void append_index(SgStatement *s, int counter);

extern void change_scalar_reference(SgStatement * s, forall_reference * fref);
extern void change_array_reference(SgStatement * s, int counter, int nameflag, int indexflag);

extern void do_shift(SgStatement * s, int counter, int * amount);
extern void do_remap(SgStatement * s, int counter);

extern void show_forall_reference(SgStatement * s);

extern int pcrc_detect_communication(detect_DIM* dad_x, detect_DIM* dad_y,
                              int* l, int* u, int* stride, int *a0, int *b0,
                              int* a1, int* b1, int *amount);

extern SgExpression * find_exp(int);
extern void tree_exp(SgExpression *, int);

void define_bounds(SgStatement * s)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();
  
  for (int i=0; i<fhdef->getindexnumber(); i++) {
    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_loop", NULL);
      SgFunctionRefExp * functionexpr = new SgFunctionRefExp (*fname);

      {
	for (int j=0; j<3; j++) {
	  if (fhdef->getindextriple(i, j)!=NULL)
	    functionexpr->addArg(*fhdef->getindextriple(i, j));
	  else {
	    SgValueExp * strideexp = new SgValueExp(1);
	    functionexpr->addArg(*strideexp);
 	  }
	}

	{
	  SgStatement * body = s->childList1(0);
	  forall_reference * fref = (forall_reference *) ((body->getAttribute(0, Forall_Ref))->getAttributeData());
	  
	  SgArrayRefExp * array = (SgArrayRefExp *) fref->getexpression();

	  int dim;

	  for (dim=0; dim<array->numberOfSubscripts(); dim++)
	    if (fref->getIndex(dim, i)!=0) {
	      break;
	    };

	  SgExpression * stride;

	  if (fref->getstride(dim, i)!=NULL) { 
	    if (fref->getstride(dim,i)->isInteger())
	      stride = new SgValueExp(fref->getstride(dim,i)->valueInteger());
	    else
	      stride = fref->getstride(dim,i);
	  }
	  else 
	    stride = new SgValueExp(0);
	  functionexpr->addArg(*stride);
	    
	  SgExpression * offset;
	  if (fref->getoffset(dim,i)!=NULL) {
	    if (fref->getoffset(dim,i)->isInteger())
	      offset = new SgValueExp(fref->getoffset(dim,i)->valueInteger());
	    else
	      offset = fref->getoffset(dim,i);
	  }
	  else 
	    offset = new SgValueExp(0);
	  functionexpr->addArg(*offset);

	};

	SgSymbol * frngname = s->getSymbolInCurrentScope(mixstring, "rng", NULL);
	SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*frngname);

	SgSymbol * distsym = s->childList1(0)->expr(0)->symbol();
	SgSymbol * distdad = s->getSymbolInCurrentScope(mixstring, "dad", distsym->identifier());
	SgExpression * dad = new SgExpression(VAR_REF, NULL, NULL, distdad);
	funcexpr->addArg(*dad);

	SgValueExp * dimention = new SgValueExp (i+1);
	funcexpr->addArg(*dimention);

	functionexpr->addArg(*funcexpr);
      }

      SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);
      SgExpression * rngexp = new SgExpression (VAR_REF, NULL, NULL, irng);

      SgAssignStmt * assign = new SgAssignStmt (*rngexp, *functionexpr);

      s->insertStmtBefore(*assign, *s->controlParent());

      
      SgSymbol * funname = s->getSymbolInCurrentScope(mixstring, "delete_range", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*funname);
     
      afuncall->addArg(* rngexp);
      s->insertStmtAfter(*afuncall, *s->controlParent());
    }

    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "loop_bounds", NULL);

      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);
      SgExpression * rngexp = new SgExpression (VAR_REF, NULL, NULL, irng);

      afuncall->addArg(*rngexp);

      SgSymbol * index=fhdef->getindexsymb(i);
      Forall_index_def * fidef = (Forall_index_def *) index->getAttribute(0, FINDEX_Def)->getAttributeData();

      SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier());
      SgExpression * lbexpr = new SgExpression(VAR_REF, NULL, NULL, lb);
      afuncall->addArg(*lbexpr);
      SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier());
      SgExpression * ubexpr = new SgExpression(VAR_REF, NULL, NULL, ub);
      afuncall->addArg(*ubexpr);
      SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier());
      SgExpression * stepexpr = new SgExpression(VAR_REF, NULL, NULL, step);
      afuncall->addArg(*stepexpr);
    
      s->insertStmtBefore(*afuncall, *s->controlParent());
    };
  };
};

void add_do_loop(SgStatement * s)
{

  SgStatement * inner;

  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  if (fhdef->getcondition()==NULL) {
    //inner=s->childList1(0);
    inner=&s->childList1(0)->copy();
  }
  else {
    //inner = new SgIfStmt(*fhdef->getcondition(), *s->childList1(0));
    inner = new SgIfStmt(*fhdef->getcondition(), s->childList1(0)->copy());
  };

  SgStatement * newbody=inner;

  int i;

  for (i=fhdef->getindexnumber(); i>0; i--) {
     SgSymbol * index=fhdef->getindexsymb(i-1);

     Forall_index_def * fidef = (Forall_index_def *) index->getAttribute(0, FINDEX_Def)->getAttributeData();

     //SgSymbol * lb = fidef->getlb();
     //SgExpression * lbexpr = new SgExpression(VAR_REF, NULL, NULL, lb);

     SgValueExp * lbexpr = new SgValueExp(0);

     SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier());
     SgExpression * ubexpr = new SgExpression(VAR_REF, NULL, NULL, ub);
     {
       SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier());
       SgExpression * lbexpr = new SgExpression(VAR_REF, NULL, NULL, lb);
       (*ubexpr) =(* ubexpr) - (* lbexpr);
       
       SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier());
       SgExpression * stepexpr = new SgExpression(VAR_REF, NULL, NULL, step);
       (*ubexpr) =(* ubexpr) / (* stepexpr);
     }

     //SgSymbol * step = fidef->getstep();
     //SgExpression * stepexpr = new SgExpression(VAR_REF, NULL, NULL, step);

     SgValueExp * stepexpr = new SgValueExp(1);
   
     SgStatement * loop = new SgForStmt(*index, *lbexpr, *ubexpr, *stepexpr, *inner);

     inner = loop;
  };

  // add those index definition statement;

  for (i=1; i<s->numberOfChildrenList1();i++) {
    SgStatement * index = &s->childList1(i)->copy();
    newbody->insertStmtBefore(*index, *newbody->controlParent());
  };

  // add procossor position detection;

  SgSymbol * grp = s->getSymbolInCurrentScope(mixstring, "grp", NULL);
  SgFunctionRefExp * grpexp = new SgFunctionRefExp(*grp);

  SgSymbol * lhssec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, 0);
  SgExpression * funref= new SgExpression(VAR_REF, NULL, NULL, lhssec);

  grpexp->addArg(*funref);

  SgSymbol * on = s->getSymbolInCurrentScope(mixstring, "on", NULL);
  SgFunctionRefExp * onexp = new SgFunctionRefExp(*on);
  
  onexp->addArg(*grpexp);

  inner = new SgIfStmt(*onexp, *inner);

  s->insertStmtBefore(*inner, *s->controlParent());
}

//new function after analyse;

void change_body(SgStatement * s)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  SgStatement * body = s->childList1(0);
  forall_reference * leftref = (forall_reference *) ((body->getAttribute(0, Forall_Ref))->getAttributeData());
  
  int counter=0;
  int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};

  add_coef(s, counter, amount);

  create_section(s, counter);

  append_index(s, counter);

  for (int i=1; i<body->numberOfAttributes(Forall_Ref); i++) {
    forall_reference * rightref = (forall_reference *) ((body->getAttribute(i, Forall_Ref))->getAttributeData());

    switch (rightref->getexpression()->variant()) {
    case VAR_REF:
      change_scalar_reference(s, rightref);
      break;
    case ARRAY_REF:
      counter++;

      switch (rightref->getCommunication()) {
      case IRREG: {
	printf("not support irreg!\n");
	{
	  printf("irregular\n");
	  for(int j=0; j<MAX_ALIGN_DIM;i++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}
	
	break;
      }

      case UNKNOW: {
	{
	  printf("unknow\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}
      
	do_remap(s, counter);
	change_array_reference(s, counter, 1,1);
	break;
      }

      case NONE: {
	{
	  printf("none\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}

	if (leftref->getexpression()->symbol()!=rightref->getexpression()->symbol()) {
	  add_coef(s, counter, amount);
	  append_index(s, counter);
	  change_array_reference(s, counter, 0,0);
	}
	else {
	  // we can do better;
	  add_coef(s, counter, amount);
	  append_index(s, counter);
	  change_array_reference(s, counter, 0,0);
	} 

	break;
      }
      
      case SHIFT: { 	    
	{
	  printf("shift\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	      printf("\n");
	};
	    
	int shiftdim=0;
	for (int j=0; j<MAX_ALIGN_DIM;j++)
	  if (rightref->getAmount(j)>0)
	    shiftdim++;
	
	if (shiftdim>1) {
	  do_remap(s, counter);
	  change_array_reference(s, counter, 1,1);
	  break;
	}
	else {
	  int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};
	  
	  for (int j=0; j<MAX_ALIGN_DIM;j++)
	    amount[j]=rightref->getAmount(j);
	  
	  add_coef(s, counter, amount);
	  
	  do_shift(s, counter, amount);
	  append_index(s, counter);
	  change_array_reference(s, counter, 1,0);
	  break;
	}
      }
      
      case REMAP: {
	{
	  printf("remap\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}
      
	do_remap(s, counter);
	change_array_reference(s, counter, 1,1);
	break;
      }
	    
      case BROADCAST: {
	{
	  printf("broadcast\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}
	printf("not supported\n");
	break;
      }

      default:
	printf("Return valuse of detect_communication is invalid!\n");
      }
    }
  }

  change_array_reference(s, 0, 0,0);

}

void trans_forall(SgStatement * s)
{
  SgStatement * body=s->childList1(0);
  forall_reference * leftref = (forall_reference *) ((body->getAttribute(0, Forall_Ref))->getAttributeData());
  
  SgExpression * refexp=leftref->getexpression();

  if (refexp->variant()==VAR_REF) {
    printf("Assignment to scalar in loop!\n");
  }
  else if (refexp->variant()==ARRAY_REF) {
    define_bounds(s);
    change_body(s);
    add_do_loop(s);	

  };

}













